"""Tests to ensure the LLaMA v2 prompt is correct."""

from kani import ChatMessage
from kani.engines.llama2_prompt import build_str


def prompt_str(messages: list[ChatMessage]) -> str:
    out = ""
    for content, bos, eos in build_str(messages):
        if bos:
            out += "<s>"
        out += content
        if eos:
            out += "</s>"
    return out


def test_basic():
    messages = [
        ChatMessage.user("Hello there."),
        ChatMessage.assistant("General Kenobi."),
    ]
    expected = "<s>[INST] Hello there. [/INST] General Kenobi. </s>"
    assert prompt_str(messages) == expected

    messages = [
        ChatMessage.system("I am a system message."),
        ChatMessage.user("Hello there."),
        ChatMessage.assistant("General Kenobi."),
    ]
    expected = "<s>[INST] <<SYS>>\nI am a system message.\n<</SYS>>\n\nHello there. [/INST] General Kenobi. </s>"
    assert prompt_str(messages) == expected


def test_2round():
    messages = [
        ChatMessage.system("I am a system message."),
        ChatMessage.user("Hello there."),
        ChatMessage.assistant("General Kenobi."),
        ChatMessage.user("I am:"),
        ChatMessage.assistant("a potato."),
    ]
    expected = (
        "<s>[INST] <<SYS>>\nI am a system message.\n<</SYS>>\n\nHello there. [/INST] General Kenobi. </s>"
        "<s>[INST] I am: [/INST] a potato. </s>"
    )
    assert prompt_str(messages) == expected


def test_2system():
    messages = [
        ChatMessage.system("I am a system message."),
        ChatMessage.system("But wait, there's more."),
        ChatMessage.user("Hello there."),
        ChatMessage.assistant("General Kenobi."),
    ]
    expected = (
        "<s>[INST] <<SYS>>\nI am a system message.\n<</SYS>>\n\n"
        "<<SYS>>\nBut wait, there's more.\n<</SYS>>\n\nHello there. [/INST] General Kenobi. </s>"
    )
    assert prompt_str(messages) == expected


def test_2user():
    messages = [
        ChatMessage.system("I am a system message."),
        ChatMessage.user("Hello there."),
        ChatMessage.user("I am another message."),
        ChatMessage.assistant("General Kenobi."),
    ]
    expected = (
        "<s>[INST] <<SYS>>\nI am a system message.\n<</SYS>>\n\n"
        "Hello there.I am another message. [/INST] General Kenobi. </s>"
    )
    assert prompt_str(messages) == expected


def test_2asst():
    messages = [
        ChatMessage.system("I am a system message."),
        ChatMessage.user("Hello there."),
        ChatMessage.assistant("General Kenobi."),
        ChatMessage.assistant("I am another message."),
    ]
    expected = (
        "<s>[INST] <<SYS>>\nI am a system message.\n<</SYS>>\n\n"
        "Hello there. [/INST] General Kenobi.I am another message. </s>"
    )
    assert prompt_str(messages) == expected



"""Example from the Customization docs.

This example shows how to override kani's chat history handler to log every message to a JSONL file.
"""

import os

from kani import Kani, chat_in_terminal
from kani.engines.openai import OpenAIEngine

api_key = os.getenv("OPENAI_API_KEY")
engine = OpenAIEngine(api_key, model="gpt-3.5-turbo")


class LogMessagesKani(Kani):
    # You can override __init__ and track kani-specific state:
    # in this example we keep track of the file we're logging to.
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.log_file = open("kani-log.jsonl", "w")

    async def add_to_history(self, message):
        await super().add_to_history(message)
        self.log_file.write(message.model_dump_json())
        self.log_file.write("\n")


ai = LogMessagesKani(engine)
if __name__ == "__main__":
    chat_in_terminal(ai)
    ai.log_file.close()



"""Common builder for the LLaMAv2-chat prompt.

This file is responsible for implementing the common non-strict ChatMessage to tokens translation, while handling
the nuance of the INST and SYS tokens as best as possible.
"""

import itertools
from typing import Callable, Iterable

from kani.models import ChatMessage, ChatRole

B_INST, E_INST = "[INST]", "[/INST]"
B_SYS, E_SYS = "<<SYS>>\n", "\n<</SYS>>\n\n"


def build(
    messages: list[ChatMessage], tokenize: Callable[[str], list[int]], bos_token_id: int = 1, eos_token_id: int = 2
) -> list[int]:
    """Build the tokens for a list of messages. `tokenize` should tokenize a str without special tokens."""
    tokens = []
    for content, bos, eos in build_str(messages):
        if bos:
            tokens.append(bos_token_id)
        tokens.extend(tokenize(content))
        if eos:
            tokens.append(eos_token_id)
    return tokens


def build_str(messages: list[ChatMessage]) -> Iterable[tuple[str, bool, bool]]:
    """Given a list of messages, yield a list of pairs of (content string, bos, eos)."""
    # combine consecutive instruction messages and non-instruction messages
    for is_inst, role_messages in itertools.groupby(
        messages, key=lambda m: m.role == ChatRole.USER or m.role == ChatRole.SYSTEM
    ):
        # get content within tags (if any)
        content = []
        for message in role_messages:
            if message.role == ChatRole.SYSTEM:
                content.append(f"{B_SYS}{message.text}{E_SYS}")
            else:
                content.append(message.text)
        # if the content is an instruction, return it wrapped in inst tags; otherwise don't
        content_str = "".join(content)
        if is_inst:
            yield f"{B_INST} {content_str} {E_INST}", True, False
        else:
            yield f" {content_str} ", False, True



from .models import ChatMessage


class FunctionCallResult:
    """A model requested a function call, and the kani runtime resolved it."""

    def __init__(self, is_model_turn: bool, message: ChatMessage):
        """
        :param is_model_turn: True if the model should immediately react; False if the user speaks next.
        :param message: The message containing the result of the function call, to add to the chat history.
        """
        self.is_model_turn = is_model_turn
        self.message = message


class ExceptionHandleResult:
    """A function call raised an exception, and the kani runtime has prompted the model with exception information."""

    def __init__(self, should_retry: bool, message: ChatMessage):
        """
        :param should_retry: Whether the model should be allowed to retry the call that caused this exception.
        :param message: The message containing details about the exception and/or instructions to retry, to add to the
            chat history.
        """
        self.should_retry = should_retry
        self.message = message



import asyncio
from typing import Literal, overload

import aiohttp
import pydantic

from .models import ChatCompletion, Completion, FunctionSpec, OpenAIChatMessage, SpecificFunctionCall
from ..httpclient import BaseClient, HTTPException, HTTPStatusException, HTTPTimeout


class OpenAIClient(BaseClient):
    """Simple HTTP client to interface with the OpenAI API.

    This client includes exponential backoff on retryable errors.
    """

    def __init__(
        self,
        api_key: str,
        http: aiohttp.ClientSession = None,
        organization: str = None,
        retry: int = 5,
        api_base: str = "https://api.openai.com/v1",
        headers: dict = None,
    ):
        """
        :param api_key: The OpenAI API key to use.
        :param http: An aiohttp session to use (creates one by default).
        :param organization: The OpenAI org ID to bill requests to (default based on api key).
        :param retry: How many times to retry a failed request (default 5).
        :param api_base: The base URL of the API to use (default OpenAI).
        :param headers: Any additional headers to send with each request.
        """
        if headers is None:
            headers = {}
        super().__init__(http)
        self.api_key = api_key
        self.organization = organization
        self.retry = retry
        self.SERVICE_BASE = api_base
        self.headers = headers

    async def request(self, method: str, route: str, headers=None, retry=None, **kwargs):
        if headers is None:
            headers = self.headers.copy()
        # set up auth headers
        if "Authorization" not in headers:
            headers["Authorization"] = f"Bearer {self.api_key}"
        if self.organization:
            headers["OpenAI-Organization"] = self.organization

        # make the request
        retry = retry if retry is not None else self.retry
        for i in range(retry):
            try:
                return await super().request(method, route, headers=headers, **kwargs)
            except (HTTPStatusException, HTTPTimeout) as e:
                # raise if out of retries
                if (i + 1) >= retry:
                    raise

                # handle http status exceptions
                # non-retryable error (see https://platform.openai.com/docs/guides/error-codes/api-errors)
                if isinstance(e, HTTPStatusException) and e.status_code in (400, 401, 403, 404):
                    raise

                # otherwise, wait and try again
                retry_sec = 2**i
                self.logger.warning(f"OpenAI returned {e}, retrying in {retry_sec} sec...")
                await asyncio.sleep(retry_sec)
        raise RuntimeError(
            "Ran out of retries but no error encountered. This should never happen; please report an issue on GitHub!"
        )

    # ==== completions ====
    @overload
    async def create_completion(
        self,
        model: str,
        prompt: str = "<|endoftext|>",
        suffix: str = None,
        max_tokens: int = 16,
        temperature: float = 1.0,
        top_p: float = 1.0,
        n: int = 1,
        logprobs: int = None,
        echo: bool = False,
        stop: str | list[str] = None,
        presence_penalty: float = 0.0,
        frequency_penalty: float = 0.0,
        best_of: int = 1,
        logit_bias: dict = None,
        user: str = None,
    ) -> Completion: ...

    async def create_completion(self, model: str, **kwargs) -> Completion:
        data = await self.post("/completions", json={"model": model, **kwargs})
        try:
            return Completion.model_validate(data)
        except pydantic.ValidationError as e:
            self.logger.exception(f"Failed to deserialize OpenAI response: {data}")
            raise HTTPException(f"Could not deserialize response: {data}") from e

    # ==== chat ====
    @overload
    async def create_chat_completion(
        self,
        model: str,
        messages: list[OpenAIChatMessage],
        functions: list[FunctionSpec] | None = None,
        function_call: SpecificFunctionCall | Literal["auto"] | Literal["none"] | None = None,
        temperature: float = 1.0,
        top_p: float = 1.0,
        n: int = 1,
        stop: str | list[str] | None = None,
        max_tokens: int | None = None,
        presence_penalty: float = 0.0,
        frequency_penalty: float = 0.0,
        logit_bias: dict | None = None,
        user: str | None = None,
    ) -> ChatCompletion: ...

    async def create_chat_completion(
        self,
        model: str,
        messages: list[OpenAIChatMessage],
        functions: list[FunctionSpec] | None = None,
        **kwargs,
    ) -> ChatCompletion:
        """Create a chat completion.

        See https://platform.openai.com/docs/api-reference/chat/create.
        """
        # transform pydantic models
        if functions:
            kwargs["functions"] = [f.model_dump(exclude_unset=True) for f in functions]
        if "function_call" in kwargs and isinstance(kwargs["function_call"], SpecificFunctionCall):
            kwargs["function_call"] = kwargs["function_call"].model_dump(exclude_unset=True)
        # call API
        data = await self.post(
            "/chat/completions",
            json={
                "model": model,
                "messages": [cm.model_dump(exclude_defaults=True, mode="json") for cm in messages],
                **kwargs,
            },
        )
        try:
            return ChatCompletion.model_validate(data)
        except pydantic.ValidationError:
            self.logger.exception(f"Failed to deserialize OpenAI response: {data}")
            raise HTTPException(f"Could not deserialize response: {data}")



from . import engines, exceptions, utils
from .ai_function import AIFunction, AIParam, ai_function
from .internal import ExceptionHandleResult, FunctionCallResult
from .kani import Kani
from .models import ChatMessage, ChatRole, FunctionCall, MessagePart
from .utils.cli import chat_in_terminal, chat_in_terminal_async

# declare that kani is also a namespace package
__path__ = __import__("pkgutil").extend_path(__path__, __name__)



"""
A couple convenience formatters to customize :meth:`.Kani.full_round_str`.

You can pass any of these functions in with, e.g., ``Kani.full_round_str(..., message_formatter=all_message_contents)``.
"""

from kani.models import ChatMessage, ChatRole


def all_message_contents(msg: ChatMessage):
    """Return the content of any message."""
    return msg.text


def assistant_message_contents(msg: ChatMessage):
    """Return the content of any assistant message; otherwise don't return anything."""
    if msg.role == ChatRole.ASSISTANT:
        return msg.text


def assistant_message_contents_thinking(msg: ChatMessage):
    """Return the content of any assistant message, and "Thinking..." on function calls."""
    if msg.role == ChatRole.ASSISTANT:
        text = msg.text
        if msg.function_call and text:
            return f"{text}\n    Thinking ({msg.function_call.name})..."
        elif msg.function_call:
            return f"Thinking ({msg.function_call.name})..."
        return text



"""Example from Advanced Usage docs.

This example shows how to use the MessagePart API to implement hidden messages, such as chain-of-thought. You can
also use MessageParts in user messages, e.g. to provide images to a multimodal model.

You should be familiar with the concept of Engines before trying to understand this code.
"""

from kani import AIFunction, ChatMessage, Kani, MessagePart, chat_in_terminal
from kani.engines.base import BaseEngine, Completion
from kani.engines.openai import OpenAIEngine


# First, define the message part that will contain the hidden chain of thought data
class ThoughtPart(MessagePart):
    # hold some string data that is the chain-of-thought
    data: str

    # when stringified, this part should be hidden from the user/any engine that does not explicitly support it
    def __str__(self):
        return ""


# Then, define an engine that can use the part we defined - we'll wrap another engine to provide a translation layer
class ChainOfThoughtEngine(BaseEngine):
    def __init__(self, engine: BaseEngine):
        self.engine = engine
        self.max_context_size = engine.max_context_size

    @staticmethod
    def translate_message(message: ChatMessage) -> ChatMessage:
        """Translate a input message into a simple string-only message to pass to the underlying engine."""
        content = ""
        # iterate over parts: if it's a thought, place its data in the content; otherwise prefix it with "Answer: "
        for part in message.parts:
            if isinstance(part, ThoughtPart):
                content += f"{part.data}\n"
            else:
                content += f"Answer: {str(part)}"
        # return the translated message
        return message.copy_with(content=content.strip())

    # === BaseEngine interface ===
    def message_len(self, message: ChatMessage) -> int:
        return self.engine.message_len(self.translate_message(message))

    async def predict(
        self, messages: list[ChatMessage], functions: list[AIFunction] | None = None, **hyperparams
    ) -> Completion:
        # translate the messages
        translated_messages = [self.translate_message(m) for m in messages]

        # generate a completion using the underlying engine given those messages
        result = await self.engine.predict(translated_messages, functions, **hyperparams)

        # parse the string-completion back into parts - with some checks to make sure the model actually output the
        # right format
        text = result.message.text
        if "Answer:" in text:
            thought, answer = text.split("Answer:", 1)
            message_with_parts = result.message.copy_with(parts=[ThoughtPart(data=thought.strip()), answer.strip()])
        else:
            message_with_parts = result.message

        # we'll print the result so you can see the parts, though a real engine wouldn't want to
        print("Model response with parts:", message_with_parts.parts)
        # and return the modified answer
        return Completion(
            message=message_with_parts,
            prompt_tokens=result.prompt_tokens,
            completion_tokens=result.completion_tokens,
        )

    # additional overrides that pass-through to underlying engine
    def function_token_reserve(self, functions):
        return self.engine.function_token_reserve(functions)

    async def close(self):
        return await self.engine.close()


# Initialize our engine as a wrapper around any underlying engine
cot_engine = ChainOfThoughtEngine(OpenAIEngine())

# Now, use kani as normal - we prompt it to answer with a thought before outputting its final answer with "Answer: "
ai = Kani(
    cot_engine,
    system_prompt=(
        "When answering the user, think step by step. Output your thoughts first, then output the final answer on its"
        ' own line, in the format "Answer: {answer}".'
    ),
)

# In this engine, we print out the parts so you can see what's going on. Normally, you would consume parts in your
# application by iterating over the returned ChatMessage, rather than the engine printing it.
#
# Try this prompt:
# Roger has 5 tennis balls. He buys 2 more cans of tennis balls. Each can has 3 tennis balls. How many tennis balls does he have now?
if __name__ == "__main__":
    chat_in_terminal(ai)



import pydantic
import pytest

from kani import ChatMessage, ChatRole, MessagePart


class TestMessagePart(MessagePart):
    def __str__(self):
        return "<TestMessagePart>"


def test_basic():
    msg = ChatMessage(role=ChatRole.USER, content="Hello world")
    assert msg.role == ChatRole.USER
    assert msg.content == "Hello world"
    assert msg.text == "Hello world"
    assert msg.parts == ["Hello world"]
    assert msg.name is None
    assert msg.function_call is None


def test_immutable():
    msg = ChatMessage(role=ChatRole.USER, content="Hello world")
    with pytest.raises(pydantic.ValidationError):
        msg.content = "not allowed"
    assert msg.content == "Hello world"


def test_none_content():
    msg = ChatMessage(role=ChatRole.USER, content=None)
    assert msg.content is None
    assert msg.text is None
    assert msg.parts == []


def test_parts():
    part = TestMessagePart()
    msg = ChatMessage(role=ChatRole.USER, content=["Hello world", part])
    assert msg.content == ("Hello world", part)
    assert msg.text == "Hello world<TestMessagePart>"
    assert msg.parts == ["Hello world", part]


def test_copy():
    part = TestMessagePart()
    msg = ChatMessage(role=ChatRole.USER, content=["Hello world", part])

    text_copy = msg.copy_with(text="asdf")
    assert text_copy.content == "asdf"

    part_copy = msg.copy_with(parts=["foo"])
    assert part_copy.content == ("foo",)

    content_copy = msg.copy_with(content="zxcv")
    assert content_copy.content == "zxcv"

    assert msg.content == ("Hello world", part)

    with pytest.raises(ValueError):
        msg.copy_with(text="foo", parts=[])



"""Save -> load should be an identity transformation."""

import random
import string

from kani import ChatMessage, Kani, MessagePart
from tests.engine import TestEngine

engine = TestEngine()


async def test_saveload_str(tmp_path):
    """Test that basic string content messages are saved."""
    # randomly initialize a kani state
    ai = Kani(engine, desired_response_tokens=3, system_prompt="1", always_included_messages=[ChatMessage.user("2")])
    for _ in range(5):
        query_len = random.randint(0, 5)
        query = "".join(random.choice(string.ascii_letters) for _ in range(query_len))
        await ai.chat_round_str(query, test_echo=True)

    # save and load
    ai.save(tmp_path / "pytest.json")
    loaded = Kani(engine)
    loaded.load(tmp_path / "pytest.json")

    # assert equality
    assert ai.always_included_messages == loaded.always_included_messages
    assert ai.chat_history == loaded.chat_history


class TestMessagePart1(MessagePart):
    data: str


class TestMessagePart2(MessagePart):
    data: str


async def test_saveload_messageparts(tmp_path):
    """Test that message parts are serialized and deserialized into the right classes."""
    apart1 = TestMessagePart1(data="apart1")
    apart2 = TestMessagePart2(data="apart2")
    hpart1 = TestMessagePart1(data="hpart1")
    hpart2 = TestMessagePart2(data="hpart2")
    # ensure that different instances with the same data are the same
    assert apart1 == TestMessagePart1(data="apart1")
    # ensure that different classes/data are not
    assert apart1 != TestMessagePart2(data="apart1")
    assert apart1 != hpart1

    # init kani state
    ai = Kani(
        engine,
        always_included_messages=[
            ChatMessage.user(["astr", apart1]),
            ChatMessage.user([apart2, "astr2"]),
        ],
        chat_history=[
            ChatMessage.user(["hstr", hpart1]),
            ChatMessage.user([hpart2, "hstr2"]),
        ],
    )

    # save and load
    ai.save(tmp_path / "pytest.json")
    loaded = Kani(engine)
    loaded.load(tmp_path / "pytest.json")

    # assert equality
    assert ai.always_included_messages == loaded.always_included_messages
    assert ai.chat_history == loaded.chat_history



"""Model-agnostic classes used to represent the chat state and function calls."""

import abc
import enum
import json
import warnings
from typing import ClassVar, Sequence, Type, TypeAlias, Union

from pydantic import BaseModel as PydanticBase, ConfigDict, model_serializer, model_validator

from .exceptions import MissingMessagePartType

# ==== constants ====
MESSAGEPART_TYPE_KEY = "__kani_messagepart_type__"  # used for serdes of MessageParts

# ==== typing ====
MessagePartType: TypeAlias = Union["MessagePart", str]  # ChatMessage.parts[*]
QueryType: TypeAlias = str | Sequence[MessagePartType]  # Kani.*_round(...)


class BaseModel(PydanticBase, abc.ABC):
    """The base class for all Kani models."""

    def copy_with(self, **new_values):
        """Make a shallow copy of this object, updating the passed attributes (if any) to new values.
        This does not validate the updated attributes!
        This is mostly just a convenience wrapper around ``.model_copy``.
        """
        return self.model_copy(update=new_values)


# ==== chat ====
class ChatRole(enum.Enum):
    """Represents who said a chat message."""

    SYSTEM = "system"
    """The message is from the system (usually a steering prompt)."""

    USER = "user"
    """The message is from the user."""

    ASSISTANT = "assistant"
    """The message is from the language model."""

    FUNCTION = "function"
    """The message is the result of a function call."""


class FunctionCall(BaseModel):
    """Represents a model's request to call a function."""

    model_config = ConfigDict(frozen=True)

    name: str
    """The name of the requested function."""

    arguments: str
    """The arguments to call it with, encoded in JSON."""

    @property
    def kwargs(self) -> dict:
        """The arguments to call the function with, with JSON decoded to a Python dict."""
        return json.loads(self.arguments)

    @classmethod
    def with_args(cls, name: str, **kwargs):
        """Create a function call with the given arguments (e.g. for few-shot prompting)."""
        return cls(name=name, arguments=json.dumps(kwargs))


class MessagePart(BaseModel, abc.ABC):
    """Base class for a part of a message.
    Engines should inherit from this class to tag substrings with metadata or provide multimodality to an engine.
    By default, if coerced to a string, will raise a warning noting that rich message part data was lost.
    For more information see :ref:`message-parts`.
    """

    model_config = ConfigDict(frozen=True)

    # ==== serdes ====
    # used for saving/loading - map qualname to messagepart type
    _messagepart_registry: ClassVar[dict[str, Type["MessagePart"]]] = {}

    def __init_subclass__(cls, **kwargs):
        """
        When a new MessagePart is defined, we need to save its type so that we can load saved JSON into the right type
        later.
        """
        super().__init_subclass__(**kwargs)
        fqn = cls.__module__ + "." + cls.__qualname__
        if fqn in cls._messagepart_registry:
            warnings.warn(
                f"The MessagePart type {fqn!r} was defined multiple times (perhaps a class is being defined in a"
                " function scope). This may cause issues when saving/loading messages with parts of this type."
            )
        cls._messagepart_registry[fqn] = cls

    @model_serializer(mode="wrap")
    def _serialize(self, nxt):
        """Extend the default serialization dict with a key recording what type it is."""
        retval = nxt(self)
        cls = type(self)
        fqn = cls.__module__ + "." + cls.__qualname__
        retval[MESSAGEPART_TYPE_KEY] = fqn
        return retval

    # noinspection PyNestedDecorators
    @model_validator(mode="wrap")
    @classmethod
    def _validate(cls, v, nxt):
        """If we are deserializing a dict with the special key, switch to the right class' validator."""
        if isinstance(v, dict) and MESSAGEPART_TYPE_KEY in v:
            fqn = v.pop(MESSAGEPART_TYPE_KEY)
            try:
                klass = cls._messagepart_registry[fqn]
            except KeyError:
                raise MissingMessagePartType(
                    fqn,
                    f"Found a MessagePart with type {fqn!r}, but the type is not defined. Maybe the type is from an"
                    " extension that has not yet been imported?",
                )
            return klass.model_validate(v)
        return nxt(v)

    # ==== entrypoints ====
    def __str__(self):
        """
        Used to define the fallback behaviour when a part is serialized to a string (e.g. via
        :attr:`.ChatMessage.text` ).
        Override this to specify the canonical string representation of your message part.

        Engines that support message parts should generally not use this, preferring to iterate over
        :attr:`.ChatMessage.parts` instead.
        """
        type_name = type(self).__name__
        warnings.warn(
            f"Message part of type {type_name!r} was coerced into a string. Rich data may not be visible to the"
            " user/model.\nDevelopers: If this warning is incorrect, please add support for this message part in your"
            f" engine or override `{type_name}.__str__()`."
        )
        return f"<{type_name} {super().__str__()}>"


class ChatMessage(BaseModel):
    """Represents a message in the chat context."""

    model_config = ConfigDict(frozen=True)

    role: ChatRole
    """Who said the message?"""

    content: str | tuple[MessagePart | str, ...] | None
    """The data used to create this message. Generally, you should use :attr:`text` or :attr:`parts` instead."""

    @property
    def text(self) -> str | None:
        """The content of the message, as a string.
        Can be None only if the message is a requested function call from the assistant.
        If the message is comprised of multiple parts, concatenates the parts.
        """
        content = self.content
        if content is None:
            return None
        elif isinstance(content, str):
            return content
        return "".join(map(str, content))

    @property
    def parts(self) -> list[MessagePart | str]:
        """The parts of the message that make up its content.
        Can be an empty tuple only if the message is a requested function call from the assistant.

        This is a read-only list; changes here will not affect the message's content. To mutate the message content,
        use :meth:`copy_with` and set ``text``, ``parts``, or ``content``.
        """
        content = self.content
        if content is None:
            return []
        elif isinstance(content, str):
            return [content]
        return list(content)

    name: str | None = None
    """The name of the user who sent the message, if set (user/function messages only)."""

    function_call: FunctionCall | None = None
    """The function requested by the model (assistant messages only)."""

    @classmethod
    def system(cls, content: str | Sequence[MessagePart | str], **kwargs):
        """Create a new system message."""
        return cls(role=ChatRole.SYSTEM, content=content, **kwargs)

    @classmethod
    def user(cls, content: str | Sequence[MessagePart | str], **kwargs):
        """Create a new user message."""
        return cls(role=ChatRole.USER, content=content, **kwargs)

    @classmethod
    def assistant(cls, content: str | Sequence[MessagePart | str] | None, **kwargs):
        """Create a new assistant message."""
        return cls(role=ChatRole.ASSISTANT, content=content, **kwargs)

    @classmethod
    def function(cls, name: str, content: str | Sequence[MessagePart | str], **kwargs):
        """Create a new function message."""
        return cls(role=ChatRole.FUNCTION, content=content, name=name, **kwargs)

    def copy_with(self, **new_values):
        """Make a shallow copy of this object, updating the passed attributes (if any) to new values.

        This does not validate the updated attributes!
        This is mostly just a convenience wrapper around ``.model_copy``.

        Only one of (content, text, parts) may be passed and will update the other two attributes accordingly.
        """
        # ensure the content is immutable
        if "content" in new_values and not isinstance(new_values["content"], str):
            new_values["content"] = tuple(new_values["content"])
        # ensure that setting either text or parts works
        if "text" in new_values:
            if "content" in new_values:
                raise ValueError("At most one of ('content', 'text', 'parts') can be set.")
            new_values["content"] = new_values.pop("text")
        if "parts" in new_values:
            if "content" in new_values:
                raise ValueError("At most one of ('content', 'text', 'parts') can be set.")
            new_values["content"] = tuple(new_values.pop("parts"))
        return super().copy_with(**new_values)
